from builtins import str
import json
import os
import sys
import traceback
import re

import boto3
from db_samvaad.db_handler import DBMethod
from botocore.exceptions import ClientError

from algowrap import constants as cnsts
from algowrap import credentials as creds
from algowrap.sentry_client import sentry_client
from algowrap import sys_utils
from algowrap import err_constants as err_cnsts


def get_user_info_iteration(iter_id):
    """
    Fetches user information for the given iteration ID from the Database using
    db_samvad.

    Parameters
    -----------
    iter_id: Integer
        Iteration ID to fetch for user info

    Returns
    -------
    Dict
        Contains user ID and email.
    """
    iter_type = DBMethod.get_value_from_iter_id(iter_id, 'type')
    if iter_type == cnsts.ITER_WORKFLOW:
        workflow_req_id = DBMethod.get_value_from_iter_id(iter_id,
                                                           cnsts.WORKFLOW_REQUEST)
        user_id = int(DBMethod.get_value('workflow_request', 'owner', workflow_req_id))
        user_email = str(DBMethod.get_value('"user"', 'email', user_id))    
    elif iter_type == cnsts.ITER_COMPONENT:
        run_id = DBMethod.get_value_from_iter_id(iter_id,
                                                  cnsts.RUN_IDFR)
        user_id = int(DBMethod.get_value('runs', '"user"', run_id))
        user_email = str(DBMethod.get_value('"user"', 'email', user_id))
    algo_id = DBMethod.get_value('iteration','algorithm',iter_id)
    algorithm = DBMethod.get_value('algorithm','name',algo_id)

    return {'user_id': str(user_id), 'email': user_email, 'algorithm': algorithm}


def get_parentid_from_iteration(iter_id):
    """
    Returns the parent id from a given iteration id. 
    Parent can be workflow request or run.

    Parameters
    ----------
    iter_id: Integer
        Iteration id
    
    Returns
    -------
    Integer
        Workflow request id or run id.
    """
    iter_type = DBMethod.get_value_from_iter_id(iter_id, 'type')
    if iter_type == cnsts.ITER_WORKFLOW:
        return DBMethod.get_value_from_iter_id(iter_id,
                                                cnsts.WORKFLOW_REQUEST)
    elif iter_type == cnsts.ITER_COMPONENT:
        return DBMethod.get_value_from_iter_id(iter_id,
                                                cnsts.RUN_IDFR)


def get_errorlog_outputkey():
    """
    Returns the storage output key for the error log file.

    Returns
    -------
    String
        Key for storing the error log file in format:
            parentid/iterid/error.log
    """
    return str(str(get_parentid_from_iteration(creds.ITER_ID)) +
           cnsts.SEPARAT + str(creds.ITER_ID) +
           cnsts.SEPARAT + cnsts.LOG_FILE_NAME)


class BaseExceptionError(Exception):
    """
    Base class for custom algowrap exceptions.
    """
    message = cnsts.BASE_ERR

    def __init__(self, message=None, status_code=None, payload=None):
        """
        Constructor for the BaseExceptionError class.

        Appends iteration id to the error message to be logged. Adds user
        information to the message by fetching from the database with the
        associated iteration. Also, generates the signed URL for error.log and
        writes error to it along with the stack trace. Finally, sends the error
        and debug info to Sentry dashboard.


        Parameters
        ----------
        message: String
            Message for the error raised
        status_code: Integer
            Status code for the error raised
        payload: String/Dict
            Error message payload which can be used to append useful information
            with the error.
        """
        Exception.__init__(self)
        
        # check if the type of message is string. If not string, convert 
        # type from Exception to string
        if type(message) != str:
            message = str(message)
            
        if message:
            error_code_pattern = re.compile("^[A-Z_0-9]+ERR[0-9]+:")
            if error_code_pattern.match(str(message)):
                self.message = str(message) + cnsts.ITER_ID + creds.ITER_ID
            else:
                self.message = err_cnsts.BASE_ERROR_CODE + str(message) + cnsts.ITER_ID + creds.ITER_ID
        if status_code:
            self.status_code = status_code
        self.payload = payload
        if not message:
            self.message = cnsts.BASE_ERR
        user_info = get_user_info_iteration(creds.ITER_ID)
        sentry_client.queue_debug_log(cnsts.USER_INFO, data=user_info)
        error_log_path = creds.MOUNT_FILE_PATH + cnsts.SEPARAT + creds.ITER_ID
        if not os.path.exists(error_log_path):
            os.makedirs(error_log_path)
        error_file_path = error_log_path + cnsts.SEPARAT + cnsts.LOG_FILE_NAME
        if message is None:
            message = err_cnsts.BASE_SOME_ERROR
        type_, value_, traceback_ = sys.exc_info()
        log_message = {'message': str(message),
                       'traceback': str(traceback.format_tb(traceback_)),
                       'type': str(type_),
                       'value': str(value_),
                       'user_info': user_info
                       }
        if user_info["email"]:
            self.message = user_info["email"] + ", " + self.message
        with open(error_file_path, 'w') as logfile:
            logfile.write(json.dumps(log_message))
        try:
            s3_client = boto3.client('s3')
            s3_client.upload_file(error_file_path, creds.ITER_OUTPUT_S3_BUCKET,
                                  get_errorlog_outputkey())
        except Exception as error:
            sentry_client.log_error(err_cnsts.LOGERR1_LOG_UPLOAD_ERR, error)
        os.remove(error_file_path)
        sentry_client.log_error(self.message)
        sentry_client.logall_debug_queue()
        sys_utils.clear_output_dirs(creds.ITER_ID)

    def __str__(self):
        return self.message


class HTTPRequestError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.LOGERR2_HTTP_REQ_ERR
        BaseExceptionError.__init__(self, message)


class S3OperationError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.S3ERR2_S3_ERR
        BaseExceptionError.__init__(self, message)


class AlgorithmError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.NACORRERR1_NA_CORR_ERR
        BaseExceptionError.__init__(self, message)


class AlgoWrapperError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.ALGWRPERR1_ALGO_WRAP_ERR
        BaseExceptionError.__init__(self, message)


class InvalidInputError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.INPUTERR16_ALGO_WRAP_ERR_INVALID_INP
        BaseExceptionError.__init__(self, message)


class PackageRunError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.NACORRERR2_CORNA_PACK_ERR
        BaseExceptionError.__init__(self, message)


class InvalidServerResponseError(BaseExceptionError):
    def __init__(self, message=None):
        if not message:
            message = err_cnsts.SVRERR30_INVALID_SERVER_RESPONSE
        BaseExceptionError.__init__(self, message)

class CurveFitError(Exception):
    pass
    